import configparser
import io
import itertools
import unittest
from typing import List, Optional

from parameterized import parameterized

import imperfect

SAMPLES = [
    "[s]\nb=1",
    "[s]\nb=1\nc=2",
    "[s]\nb=1\n\n",
]

# This is for future adaptation to hypothesis, perhaps
def variations(section: Optional[str], k: str, v: str) -> List[str]:
    assert section
    section_strs = [f"\n\n[{section}]\n", f"[{section}]\n"]
    return [
        "".join(c)
        for c in itertools.product(
            section_strs,
            ["", "\n"],
            ["", " ", "  ", "\t", "    \t     "],
            [k],
            ["", " "],
            ["=", ":"],
            ["", " "],
            [v],
            ["", "\n", "\n\n"],
        )
    ]


class ImperfectTests(unittest.TestCase):
    # The goal is to get to 100% test coverage with these, but then also have
    # the hypothesis-based tests that validate behavior on many invented
    # examples.
    @parameterized.expand(  # type: ignore
        [("[s]\na=1",), ("[s]\na = 1",), ("[s]\na = 1\n",), ("[s]\na=\n  1",),],
    )
    def test_simple_parse(self, example: str) -> None:
        conf = imperfect.parse_string(example)
        # Minimal mapping api
        self.assertEqual(["s"], conf.keys())
        self.assertTrue("s" in conf)
        self.assertEqual(["a"], conf["s"].keys())
        self.assertTrue("a" in conf["s"])

        # KeyError coverage
        self.assertFalse("b" in conf)
        self.assertFalse("b" in conf["s"])

        self.assertIn(conf["s"]["a"], ("1", "\n1"))

    @parameterized.expand([("a=1",),],)  # type: ignore
    def test_fail_to_parse(self, example: str) -> None:
        with self.assertRaises(imperfect.ParseError):
            imperfect.parse_string(example)

    def test_multiline_with_comment(self) -> None:
        conf = imperfect.parse_string("[s]\na=\n #comment\n b\n")
        self.assertEqual("\nb", conf["s"]["a"])

    def test_allow_no_value(self) -> None:
        conf = imperfect.parse_string("[s]\na=", allow_no_value=True)
        self.assertEqual("", conf["s"]["a"])

    def test_alternate_delimiters(self) -> None:
        conf = imperfect.parse_string("[s]\naqq1", delimiters=("qq",))
        self.assertEqual("1", conf["s"]["a"])

    def test_alternate_delimiters_allow_no_value(self) -> None:
        conf = imperfect.parse_string(
            "[s]\naqq", delimiters=("qq",), allow_no_value=True
        )
        self.assertEqual("", conf["s"]["a"])

    def test_comment_prefixes(self) -> None:
        conf = imperfect.parse_string("[s]\n@comment\na=1", comment_prefixes=("@",))
        self.assertEqual("1", conf["s"]["a"])

    @parameterized.expand(  # type: ignore
        [
            ("[ ]=",),
            (" [0]",),
            ("[s]\na=1",),
            ("[s]\n[s2]\na=1",),
            ("[s]\n  a = 1  \n\n",),
            ("#comment\n[s]\na=1\n#comment2",),
        ],
        name_func=(lambda a, b, c: f"{a.__name__}_{b}"),
    )
    def test_simple_roundtrips(self, example: str) -> None:
        conf = imperfect.parse_string(example)
        buf = io.StringIO()
        conf.build(buf)
        self.assertEqual(buf.getvalue(), example)

    def test_exhaustive_roundtrip(self) -> None:
        for example in variations("sect", "a", "1"):
            oracle = configparser.RawConfigParser()
            try:
                oracle.read_string(example)
                self.assertEqual(oracle["sect"]["a"], "1")
            except Exception:  # pragma: no cover
                continue

            conf = None
            try:
                conf = imperfect.parse_string(example)
                # Did we parse the same value as configparser
                self.assertEqual(len(conf["sect"].entries), 1)
                self.assertEqual(conf["sect"].entries[0].key, "a")
                self.assertEqual(conf["sect"].entries[0].interpret_value(), "1")
            except Exception:  # pragma: no cover
                print("Input: ", repr(example))
                if conf:
                    # if conf.default:
                    #     print("default:")
                    #     for v in conf.default.entries:
                    #         print("  ", v)
                    for s in conf.sections:
                        print(f"{s.name}:")
                        for v in s.entries:
                            print("  ", v)
                raise

            buf = io.StringIO()
            conf.build(buf)
            self.assertEqual(buf.getvalue(), example)

    def test_example_from_readme(self) -> None:
        INPUT = """
[metadata]
# the package name
name = imperfect
# slurp the readme
long_description = file: README.md

[options]
packages = imperfect
"""
        EXPECTED = """
[metadata]
# the package name
name = imperfect
# slurp the readme
long_description = file: README.md
long_description_content_type =  text/markdown

[options]
packages = imperfect
"""
        import io
        import difflib
        import imperfect

        data = INPUT

        conf: imperfect.ConfigFile = imperfect.parse_string(data)
        metadata_section = conf["metadata"]

        # Ignoring some whitespace for now, this looks like
        # long_description_content_type =  text/markdown\n
        # [                   entry                      ]
        # [            key            ][eq][    value    ]

        value = imperfect.ValueLine(
            whitespace_before_text="",
            text="text/markdown",
            whitespace_after_text="",
            newline="\n",
        )
        new_entry = imperfect.ConfigEntry(
            key="long_description_content_type",
            equals="=",
            value=[value],
            whitespace_before_equals=" ",
            whitespace_before_value="  ",
        )
        for i, entry in enumerate(metadata_section.entries):
            if entry.key == "long_description":
                metadata_section.entries.insert(i + 1, new_entry)
                break

        buf = io.StringIO()
        conf.build(buf)

        self.assertEqual(EXPECTED, buf.getvalue())

        temp = "".join(
            difflib.unified_diff(
                data.splitlines(True), buf.getvalue().splitlines(True),
            )
        )
        self.assertTrue(temp)



from dataclasses import dataclass, field
from typing import Any, List


class ParseError(Exception):
    pass


@dataclass
class ConfigFile:
    sections: List["ConfigSection"] = field(default_factory=list)
    # The naming of these comes from configobj
    initial_comment: str = ""
    final_comment: str = ""

    def keys(self) -> List[str]:
        return [s.name for s in self.sections]

    def __getitem__(self, name: str) -> "ConfigSection":
        for s in self.sections:
            if s.name == name:
                return s
        raise KeyError(f"Missing section {name}")

    def __contains__(self, name: str) -> bool:
        try:
            self[name]
            return True
        except KeyError:
            return False

    # TODO file-like type
    def build(self, buf: Any) -> None:
        buf.write(self.initial_comment)
        for s in self.sections:
            s.build(buf)
        buf.write(self.final_comment)

    def set_value(self, section: str, key: str, value: str) -> None:
        try:
            s = self[section]
        except KeyError:
            s = ConfigSection(
                leading_square_bracket="[",
                name=section,
                trailing_square_bracket="]",
                newline="\n",
                leading_whitespace="",
                trailing_whitespace="",
            )
            self.sections.append(s)
        s.set_value(key, value)


@dataclass
class ConfigSection:
    leading_whitespace: str
    leading_square_bracket: str
    name: str
    trailing_square_bracket: str
    trailing_whitespace: str
    newline: str
    entries: List["ConfigEntry"] = field(default_factory=list)

    # TODO file-like type
    def build(self, buf: Any) -> None:
        buf.write(
            self.leading_whitespace
            + self.leading_square_bracket
            + self.name
            + self.trailing_square_bracket
            + self.trailing_whitespace
            + self.newline
        )
        for e in self.entries:
            e.build(buf)

    def keys(self) -> List[str]:
        return [e.key.lower() for e in self.entries]

    def __getitem__(self, name: str) -> str:
        for e in self.entries:
            if e.key.lower() == name.lower():
                return e.interpret_value()
        raise KeyError(name)

    def __contains__(self, name: str) -> bool:
        try:
            self[name]
            return True
        except KeyError:
            return False

    def set_value(self, key: str, value: str) -> None:
        new_valueline = ValueLine(
            text=value,
            newline="\n",
            whitespace_before_text="",
            whitespace_after_text="",
        )
        for e in self.entries:
            if e.key.lower() == key:
                e.value = [new_valueline]
                break
        else:
            self.entries.append(
                ConfigEntry(
                    key=key,
                    equals="=",
                    value=[new_valueline],
                    whitespace_before_equals=" ",
                    whitespace_before_value=" ",
                )
            )


@dataclass
class ConfigEntry:
    key: str
    equals: str
    value: List["ValueLine"] = field(default_factory=list)

    whitespace_before_key: str = ""
    whitespace_before_equals: str = ""
    whitespace_before_value: str = ""
    whitespace_after_value: str = ""  # The final (though optional) newline

    def interpret_value(self) -> str:
        return "".join(
            [
                v.text + (v.newline if i < (len(self.value) - 1) else "")
                for i, v in enumerate(self.value)
            ]
        )

    # TODO
    def build(self, buf: Any) -> None:
        buf.write(
            self.whitespace_before_key
            + self.key
            + self.whitespace_before_equals
            + self.equals
            + self.whitespace_before_value
        )
        for v in self.value:
            v.build(buf)
        buf.write(self.whitespace_after_value)


@dataclass
class ValueLine:
    whitespace_before_text: str
    text: str
    whitespace_after_text: str
    newline: str

    # TODO
    def build(self, buf: Any) -> None:
        buf.write(
            self.whitespace_before_text
            + self.text
            + self.whitespace_after_text
            + self.newline
        )



import io
import unittest

from .. import parse_string


class EditingTest(unittest.TestCase):
    def test_existing_section_change_entry_preserves_space(self) -> None:
        # Space around the '=' is not changed when you change the value
        conf = parse_string("[a]\n#comment1\nb = 1\n#comment2\n")
        conf.set_value("a", "b", "2")
        buf = io.StringIO()
        conf.build(buf)
        self.assertEqual("[a]\n#comment1\nb = 2\n#comment2\n", buf.getvalue())

    def test_existing_section_change_entry(self) -> None:
        conf = parse_string("[a]\n#comment1\nb=1\n#comment2\n")
        conf.set_value("a", "b", "2")
        buf = io.StringIO()
        conf.build(buf)
        self.assertEqual("[a]\n#comment1\nb=2\n#comment2\n", buf.getvalue())

    def test_existing_section_new_entry(self) -> None:
        conf = parse_string("[a]\nb = 1\n")
        conf.set_value("a", "c", "2")
        buf = io.StringIO()
        conf.build(buf)
        self.assertEqual("[a]\nb = 1\nc = 2\n", buf.getvalue())

    def test_new_section_new_entry(self) -> None:
        conf = parse_string("")
        conf.set_value("a", "b", "1")
        buf = io.StringIO()
        conf.build(buf)
        self.assertEqual("[a]\nb = 1\n", buf.getvalue())



import string
import unittest
from configparser import RawConfigParser
from typing import Any, Callable, List, Union

import hypothesis.strategies as st
from hypothesis import HealthCheck, example, given, settings

import imperfect

whitespace = (" ", "\t", "\n", "\r")
line_noise = ("#", ";", "=", ":")
letters = ("a", "b", "c")
chars = st.characters()


@st.composite
def simple_well_formed_ini(draw: Callable[[Any], str]) -> str:
    section_name = draw(st.text())
    item_name = draw(st.text())
    item_value = draw(st.text())
    template = "x[x{section_name}x]xx{item_name}x=x{item_value}x"
    # replace all the 'x' with a random whitespace
    while True:
        i = template.find("x")
        if i == -1:
            break
        template = template[:i] + draw(st.text(alphabet=" \t\r\n")) + template[i + 1 :]
    return template.format(**locals())


@st.composite
def ini_section(draw: Callable[[Any], str]) -> str:
    section_name = draw(st.text(alphabet=chars))
    template = "x[x{section_name}x]x"
    while True:
        i = template.find("x")
        if i == -1:
            break
        template = (
            template[:i] + draw(st.text(alphabet=string.whitespace)) + template[i + 1 :]
        )
    return template.format(**locals())


@st.composite
def ini_value(draw: Callable[[Any], str]) -> str:
    a = draw(st.text(alphabet=chars))
    b = draw(st.text(alphabet=chars))
    c = draw(st.text(alphabet=chars))
    eq = draw(st.text(alphabet="=:"))
    template = "x{a}x{eq}x{b}x{c}x"
    while True:
        i = template.find("x")
        if i == -1:
            break
        template = (
            template[:i] + draw(st.text(alphabet=string.whitespace)) + template[i + 1 :]
        )
    return template.format(**locals())


@st.composite
def continued_value(draw: Callable[[Any], str]) -> str:
    whitespace = draw(st.text(alphabet=string.whitespace))
    value = draw(st.text())
    return whitespace + value


def configparser_is_ok_with_it(**kwargs: Any) -> Callable[..., bool]:
    def inner(text: Union[List[str], str]) -> bool:
        if isinstance(text, list):
            text = "\n".join(text)

        try:
            RawConfigParser(strict=True, **kwargs).read_string(text)
            return True
        except Exception:
            return False

    return inner


class ImperfectHypothesisTest(unittest.TestCase):
    @given(simple_well_formed_ini().filter(configparser_is_ok_with_it()))
    @example("[a]\nb=1")
    @settings(max_examples=1000, deadline=10)  # type: ignore
    def atest_parse(self, text: str) -> None:
        print("Validating", repr(text))
        rcp = RawConfigParser(strict=True)
        rcp.read_string(text)

        conf = imperfect.parse_string(text)

        for section in rcp:
            if section != "DEFAULT":
                self.assertIn(section, conf)  # type: ignore
                self.assertEqual(
                    conf[section].keys(), list(rcp[section].keys()),
                )

    @given(
        st.lists(ini_section() | ini_value() | continued_value(), min_size=2).filter(
            configparser_is_ok_with_it()
        )
    )
    @settings(  # type: ignore
        suppress_health_check=[HealthCheck.too_slow], max_examples=200, deadline=3000
    )
    @example(["[\r]", " 0=1"])
    @example(["[0]", " 0=", " [ ]"])
    @example(["[ ]", " 0=", " [\r]"])
    @example(["[ ]", "0=", "[  ]", " [0]"])
    @example(["[ ]", "0=", "  []", " ="])
    @example(["[ ]", "0=", "  =", " ="])
    @example(["[ ]", "0=", "[0]", " [\r]"])
    def test_parse(self, lines: List[str]) -> None:
        text = "\n".join(lines)
        rcp = RawConfigParser(strict=True)
        rcp.read_string(text)
        print("Validating", repr(text))
        print("  ", {k: dict(rcp[k]) for k in rcp if k != "DEFAULT"})

        conf = imperfect.parse_string(text)

        for section in rcp:
            if section != "DEFAULT":
                self.assertIn(section, conf)  # type: ignore
                self.assertEqual(
                    conf[section].keys(), list(rcp[section].keys()),
                )

    # @given(
    #     st.lists(ini_section() | ini_value() | continued_value(), min_size=2).filter(
    #         configparser_is_ok_with_it(empty_lines_in_values=False)
    #     )
    # )
    # @settings(
    #     suppress_health_check=[HealthCheck.too_slow], max_examples=200, deadline=3000
    # )
    # @example(['[ ]', '0:', '', ' [0]'])
    # def test_parse_no_empty_lines_in_values(self, lines):
    #     text = "\n".join(lines)
    #     rcp = RawConfigParser(strict=True, empty_lines_in_values=False)
    #     rcp.read_string(text)
    #     print("Validating", repr(text))
    #     print("  ", {k: dict(rcp[k]) for k in rcp if k != "DEFAULT"})

    #     conf = imperfect.parse_string(text, empty_lines_in_values=False)

    #     for section in rcp:
    #         if section != "DEFAULT":
    #             self.assertIn(section, conf)
    #             self.assertEqual(
    #                 conf[section].keys(), list(rcp[section].keys()),
    #             )
